@@ -26,7 +26,6 @@
 /* An item in the connection queue. */
 enum conn_queue_item_modes {
     queue_new_conn,   /* brand new connection. */
-    queue_redispatch, /* redispatching from side thread */
 };
 typedef struct conn_queue_item CQ_ITEM;
 struct conn_queue_item {
@@ -90,7 +89,7 @@ static pthread_mutex_t init_lock;
 static pthread_cond_t init_cond;
 
 
-static void thread_libevent_process(int fd, short which, void *arg);
+static void thread_libevent_process(evutil_socket_t fd, short which, void *arg);
 
 /* item_lock() must be held for an item before any modifications to either its
  * associated hash bucket, or the structure itself.
@@ -205,6 +204,7 @@ void stop_threads(void) {
     if (settings.verbose > 0)
         fprintf(stderr, "asking workers to stop\n");
     buf[0] = 's';
+    pthread_mutex_lock(&worker_hang_lock);
     pthread_mutex_lock(&init_lock);
     init_count = 0;
     for (i = 0; i < settings.num_threads; i++) {
@@ -216,6 +216,8 @@ void stop_threads(void) {
     wait_for_thread_registration(settings.num_threads);
     pthread_mutex_unlock(&init_lock);
 
+    // All of the workers are hung but haven't done cleanup yet.
+
     if (settings.verbose > 0)
         fprintf(stderr, "asking background threads to stop\n");
 
@@ -237,6 +239,17 @@ void stop_threads(void) {
     if (settings.verbose > 0)
         fprintf(stderr, "stopped idle timeout thread\n");
 
+    // Close all connections then let the workers finally exit.
+    if (settings.verbose > 0)
+        fprintf(stderr, "closing connections\n");
+    conn_close_all();
+    pthread_mutex_unlock(&worker_hang_lock);
+    if (settings.verbose > 0)
+        fprintf(stderr, "reaping worker threads\n");
+    for (i = 0; i < settings.num_threads; i++) {
+        pthread_join(threads[i].thread_id, NULL);
+    }
+
     if (settings.verbose > 0)
         fprintf(stderr, "all background threads stopped\n");
 
@@ -407,12 +420,38 @@ static void setup_thread(LIBEVENT_THREAD *me) {
         exit(EXIT_FAILURE);
     }
 
-    me->suffix_cache = cache_create("suffix", SUFFIX_SIZE, sizeof(char*),
-                                    NULL, NULL);
-    if (me->suffix_cache == NULL) {
-        fprintf(stderr, "Failed to create suffix cache\n");
+    me->resp_cache = cache_create("resp", sizeof(mc_resp), sizeof(char *), NULL, NULL);
+    if (me->resp_cache == NULL) {
+        fprintf(stderr, "Failed to create response cache\n");
         exit(EXIT_FAILURE);
     }
+    // Note: we were cleanly passing in num_threads before, but this now
+    // relies on settings globals too much.
+    if (settings.resp_obj_mem_limit) {
+        int limit = settings.resp_obj_mem_limit / settings.num_threads;
+        if (limit < sizeof(mc_resp)) {
+            limit = 1;
+        } else {
+            limit = limit / sizeof(mc_resp);
+        }
+        cache_set_limit(me->resp_cache, limit);
+    }
+
+    me->rbuf_cache = cache_create("rbuf", READ_BUFFER_SIZE, sizeof(char *), NULL, NULL);
+    if (me->rbuf_cache == NULL) {
+        fprintf(stderr, "Failed to create read buffer cache\n");
+        exit(EXIT_FAILURE);
+    }
+    if (settings.read_buf_mem_limit) {
+        int limit = settings.read_buf_mem_limit / settings.num_threads;
+        if (limit < READ_BUFFER_SIZE) {
+            limit = 1;
+        } else {
+            limit = limit / READ_BUFFER_SIZE;
+        }
+        cache_set_limit(me->rbuf_cache, limit);
+    }
+
 #ifdef EXTSTORE
     me->io_cache = cache_create("io", sizeof(io_wrap), sizeof(char*), NULL, NULL);
     if (me->io_cache == NULL) {
@@ -466,12 +505,12 @@ static void *worker_libevent(void *arg) {
  * Processes an incoming "handle a new connection" item. This is called when
  * input arrives on the libevent wakeup pipe.
  */
-static void thread_libevent_process(int fd, short which, void *arg) {
+static void thread_libevent_process(evutil_socket_t fd, short which, void *arg) {
     LIBEVENT_THREAD *me = arg;
     CQ_ITEM *item;
     char buf[1];
     conn *c;
-    unsigned int timeout_fd;
+    unsigned int fd_from_pipe;
 
     if (read(fd, buf, 1) != 1) {
         if (settings.verbose > 0)
@@ -518,10 +557,6 @@ static void thread_libevent_process(int fd, short which, void *arg) {
 #endif
                 }
                 break;
-
-            case queue_redispatch:
-                conn_worker_readd(item->c);
-                break;
         }
         cqi_free(item);
         break;
@@ -531,12 +566,21 @@ static void thread_libevent_process(int fd, short which, void *arg) {
         break;
     /* a client socket timed out */
     case 't':
-        if (read(fd, &timeout_fd, sizeof(timeout_fd)) != sizeof(timeout_fd)) {
+        if (read(fd, &fd_from_pipe, sizeof(fd_from_pipe)) != sizeof(fd_from_pipe)) {
             if (settings.verbose > 0)
                 fprintf(stderr, "Can't read timeout fd from libevent pipe\n");
             return;
         }
-        conn_close_idle(conns[timeout_fd]);
+        conn_close_idle(conns[fd_from_pipe]);
+        break;
+    /* a side thread redispatched a client connection */
+    case 'r':
+        if (read(fd, &fd_from_pipe, sizeof(fd_from_pipe)) != sizeof(fd_from_pipe)) {
+            if (settings.verbose > 0)
+                fprintf(stderr, "Can't read redispatch fd from libevent pipe\n");
+            return;
+        }
+        conn_worker_readd(conns[fd_from_pipe]);
         break;
     /* asked to stop */
     case 's':
@@ -591,47 +635,26 @@ void dispatch_conn_new(int sfd, enum conn_states init_state, int event_flags,
  * Re-dispatches a connection back to the original thread. Can be called from
  * any side thread borrowing a connection.
  */
+#define REDISPATCH_MSG_SIZE (1 + sizeof(int))
 void redispatch_conn(conn *c) {
-    CQ_ITEM *item = cqi_new();
-    char buf[1];
-    if (item == NULL) {
-        /* Can't cleanly redispatch connection. close it forcefully. */
-        c->state = conn_closed;
-        close(c->sfd);
-        return;
-    }
+    char buf[REDISPATCH_MSG_SIZE];
     LIBEVENT_THREAD *thread = c->thread;
-    item->sfd = c->sfd;
-    item->init_state = conn_new_cmd;
-    item->c = c;
-    item->mode = queue_redispatch;
 
-    cq_push(thread->new_conn_queue, item);
-
-    buf[0] = 'c';
-    if (write(thread->notify_send_fd, buf, 1) != 1) {
-        perror("Writing to thread notify pipe");
+    buf[0] = 'r';
+    memcpy(&buf[1], &c->sfd, sizeof(int));
+    if (write(thread->notify_send_fd, buf, REDISPATCH_MSG_SIZE) != REDISPATCH_MSG_SIZE) {
+        perror("Writing redispatch to thread notify pipe");
     }
 }
 
 /* This misses the allow_new_conns flag :( */
 void sidethread_conn_close(conn *c) {
-    c->state = conn_closed;
     if (settings.verbose > 1)
-        fprintf(stderr, "<%d connection closed from side thread.\n", c->sfd);
-#ifdef TLS
-    if (c->ssl) {
-        c->ssl_wbuf = NULL;
-        SSL_shutdown(c->ssl);
-        SSL_free(c->ssl);
-    }
-#endif
-    close(c->sfd);
-
-    STATS_LOCK();
-    stats_state.curr_conns--;
-    STATS_UNLOCK();
+        fprintf(stderr, "<%d connection closing from side thread.\n", c->sfd);
 
+    c->state = conn_closing;
+    // redispatch will see closing flag and properly close connection.
+    redispatch_conn(c);
     return;
 }
 
@@ -741,7 +764,7 @@ enum delta_result_type add_delta(conn *c, const char *key,
 
     hv = hash(key, nkey);
     item_lock(hv);
-    ret = do_add_delta(c, key, nkey, incr, delta, buf, cas, hv);
+    ret = do_add_delta(c, key, nkey, incr, delta, buf, cas, hv, NULL);
     item_unlock(hv);
     return ret;
 }
@@ -820,6 +843,11 @@ void threadlocal_stats_aggregate(struct thread_stats *stats) {
                 threads[ii].stats.lru_hits[sid];
         }
 
+        stats->response_obj_bytes += threads[ii].resp_cache->total * sizeof(mc_resp);
+        stats->response_obj_total += threads[ii].resp_cache->total;
+        stats->response_obj_free += threads[ii].resp_cache->freecurr;
+        stats->read_buf_bytes += threads[ii].rbuf_cache->total * READ_BUFFER_SIZE;
+        stats->read_buf_bytes_free += threads[ii].rbuf_cache->freecurr * READ_BUFFER_SIZE;
         pthread_mutex_unlock(&threads[ii].stats.mutex);
     }
 }
